/* wordgame -- GPLv3 -- 2010 -- pancake@nopcode.org */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>

//#define DICT "/usr/share/dict/usa"
#define DICTDIR PREFIX"/share/wg/"
#define DICT "usa"
//catalan"
#define MINLEN 3
#define MAXLEN 6
#define MAXWRD 16
#define OTIME (60*2)

// term.c ## TODO: simplify
#include <termios.h>

static struct termios _termbuf, _termraw;

static void terminit() {
	tcgetattr (0, &_termbuf);
	memcpy (&_termraw, &_termbuf, sizeof (_termbuf));
	_termraw.c_iflag &= ~(BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
	_termraw.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN|CSIZE|PARENB);
	_termraw.c_cflag |= CS8;
	_termraw.c_cc[VMIN] = 1;
}

static void termraw() {
	tcsetattr (0, TCSANOW, &_termraw);
}

static void termbuf() {
	tcsetattr (0, TCSANOW, &_termbuf);
}
// term

static char *dict;
static char curword[MAXLEN+1];
static int ncurword = 0;
static char *words[MAXLEN+1][MAXWRD];
static int nwords[MAXLEN+1];
static int wordset[MAXLEN+1];
static int tstamp = OTIME;
static char chars[MAXLEN+1];
static char chs[MAXLEN+1];
static int charset;
static int win;
static int nwins = 0;
static int score = 0;
static int view = 0;

static void appendword(char *word, int len) {
	char mychars[MAXLEN+1];
	char *ptr;
	int i;
	if (len<MINLEN || len>MAXLEN)
		return;
	if (nwords[len] == MAXWRD)
		return;
	memcpy (mychars, chs, sizeof (chs));
	for (i=0; i<len; i++) {
		word[i] = tolower (word[i]);
		ptr = strchr (mychars, word[i]);
		if (ptr) *ptr = '_';
		else return;
	}
	words[len][nwords[len]] = strdup (word);
	nwords[len]++;
}

static void shuffle() {
	int i, bit;
	char newchars[MAXLEN];
	charset = 0;
	for (i=0; i<MAXLEN; i++) {
		do bit = rand ()%MAXLEN;
		while (charset&(1<<bit));
		charset |= (1<<bit);
		newchars[i] = chars[bit];
	}
	charset = 0;
	memcpy (chars, newchars, MAXLEN);
	memcpy (chs, chars, MAXLEN);
}

static int gencharset() {
	int i;
	const char *vowels = "aeiou";
	for (i=0; i<MAXLEN; i++) {
		if (!(rand ()%5))
			chars[i] = vowels[rand ()%5];
		else chars[i] = 'a'+(rand ()%('z'-'a'));
	}
	chars[i] = '\0';
	charset = 0;
	memcpy (chs, chars, sizeof (chs));
}

static void setupgame() {
	int i;
	printf ("\x1b[2J");
	for (i=0; i<MAXLEN; i++) {
		nwords[i] = 0;
		words[i][0] = NULL;
		wordset[i] = 0;
	}
	curword[0] = ' ';
	gencharset ();
}

static void drawscreen() {
	int i, j, k;
	if (tstamp<=0) {
		tstamp = 0;
		view = 1;
	}
	printf ("\x1b[0;0H");
	printf ("= [Q]uit  [S]huffle  [R]edraw  [N]ew  [V]iew\n");
	printf ("- dict: %s   score: %d   ", dict, score);
	printf ("time: %02d:%02d  %s\n", (tstamp/60), (tstamp%60),
		(tstamp==0)?"TIME OVER":(tstamp<30)?"!!!":"");
	printf ("\n");
	printf (" [ ");
	for (i=0; i<MAXLEN; i++)
		printf ("%c ", chs[i]);
	printf ("]   ");
	switch (win) {
	case 0:
		break;
	case -1:
	case -2:
		printf ("oops :(");
		win++;
		break;
	case 1:
	case 2:
		printf ("GOOD! :D");
		win--;
		break;
	case 3:
	case 4:
		printf ("VERY GOOD");
		win--;
		break;
	case 5:
	case 6:
		printf ("NEXT LEVEL");
		win--;
		break;
	default:
		win--;
		break;
	}
	printf ("        \n [ ");
	for (i=0; i<MAXLEN; i++)
		printf ("%c ", (i<ncurword)?curword[i]:'_');
	//(charset&(1<<i))?chars[i]:'_');
	printf ("]\n");
	for (i=MINLEN; i<=MAXLEN; i++) {
		if (nwords[i]==0)
			continue;
		printf ("\n  %d: ", i);
		for (j=0;j<nwords[i];j++) {
			if (view || wordset[i] & (1<<j)) {
				printf ("%s ", words[i][j]);
			} else {
				for (k=0;k<i;k++)
					printf ("_");
				printf (" ");
			}
			if (!((j+1)%7))
				printf ("\n     ");
		}
		if (((j+1)%7))
			printf ("\n");
	}
	printf ("\n");
	fflush (stdout);
	view = 0;
	if (tstamp<1) {
		printf ("\n\nGAME OVER\n\nscore: ");
		fflush (stdout);
		fprintf (stderr, "%d\n", score);
		termbuf ();
		exit (0);
	}
}

static void hup(int sig) {
	tstamp--;
	drawscreen ();
	alarm (1);
}

static void bye(int sig) {
	printf ("bye!\n");
	fflush (stdout);
	termbuf ();
	exit (0);
}

static int checkeog() {
	int ret = 1;
	int i, j;
	for (i=MINLEN; i<=MAXLEN; i++) {
		for (j=0; j<nwords[i]; j++) {
			if (!(wordset[i]&(1<<j))) {
				ret = 0;
				break;
			}
		}
	}
	return ret;
}

static int load(const char *dict) {
	int len;
	char buf[128];
	FILE *fd = fopen (dict, "r");
	if (fd == NULL) {
		chdir (DICTDIR);
		fd = fopen (dict, "r");
		if (fd == NULL) {
			printf ("Cannot open dict %s\n", dict);
			termbuf ();
			exit (1);
		}
	}
	signal (SIGALRM, SIG_IGN);
	while (!feof (fd)) {
		if (!fgets (buf, sizeof (buf), fd))
			break;
		len = strlen (buf);
		buf[--len] = '\0';
		if (strchr (buf, '\''))
			continue;
		appendword (buf, len);
	}
	signal (SIGALRM, hup);
	fclose (fd);
	return 1;
}

static void newgame() {
	int max = 64;
	int i, nowords = 1;
	do {
		gencharset ();
		setupgame ();
		load (dict);
		for (i=0;i<MAXLEN;i++) {
			if (nwords[i]>0) {
				nowords = 0;
				break;
			}
		}
		if (max--<0) {
			printf ("No more valid words dict..\n");
			exit (1);
		}
	} while (nowords);
}

static void check() {
	int i, j;
	win = -2;
	for (i=MINLEN; i<=MAXLEN; i++) {
		for (j=0; j<nwords[i]; j++) {
			if (wordset[i] & (1<<j))
				continue;
			if (!strcmp (words[i][j], curword)) {
				score += 20;
				wordset[i] |= (1<<j);
				win = 2;
				nwins++;
				break;
			}
		}
	}
	// TODO: check finish of the game
	if (win == -2)
		nwins--;
	if (nwins>2) {
		win = 4;
		score += 37;
		tstamp += 20;
		nwins = 0;
	}
	if (nwins<0)
		nwins = 0;
	if (checkeog()) {
		win = 6;
		newgame ();
	}
	
	ncurword = 0;
	curword[0] = '_';
	charset = 0;
	memcpy (chs, chars, sizeof (chs));
}

static void play() {
	int i, ret;
	char *ptr, ch;
	signal (SIGINT, bye);
	signal (SIGKILL, bye);
	alarm (1);
	for (;;) {
		drawscreen ();
		if (read (0, &ch, 1) != 1) {
			sleep (1);
			continue;
		}
		switch (ch) {
		case 'Q':
			bye (0);
			return;
		case 'S':
			shuffle ();
			break;
		case 'R':
			printf ("\x1b[2J");
			break;
		case 'V':
			view = 1;
			break;
		case 'N':
			newgame ();
			break;
		case 8:
			// TODO: backspace
			if (ncurword>0) {
				char *ptr, ch;
				ncurword--;
				ch = curword[ncurword];
				ptr = strchr (chars, ch);
				if (ptr) {
					int delta = (int)(size_t)(ptr-chars);
					chs[delta] = ch;
					charset &= ~(1<<delta);
				} // else OOPS
				curword[ncurword] = '\0';
			}
			break;
		case '\r':
		case '\n':
			check ();
			ncurword = 0;
			break;
		default:
			ptr = strchr (chs, ch);
			if (ptr) {
				int bit = ((int)(size_t)(ptr-chs));
				charset |= (1<<bit);
				chs[bit] = '_';
				if (ncurword<=MAXLEN) {
					curword[ncurword++] = chars[bit];
					curword[ncurword] = 0;
				}
			}
			break;
		}
	}
}

int main(int argc, char **argv) {
	if (argc > 1)
		dict = argv[1];
	else dict = DICT;
	srand (getpid ());
	terminit ();
	newgame ();
	termraw ();
	play ();
	termbuf ();
	return 0;
}
